@@ -579,7 +579,7 @@ static int td_io_prep(struct thread_data *td, struct io_u *io_u, int read)
 	return 0;
 }
 
-static void put_io_u(struct thread_data *td, struct io_u *io_u)
+void put_io_u(struct thread_data *td, struct io_u *io_u)
 {
 	list_del(&io_u->list);
 	list_add(&io_u->list, &td->io_u_freelist);
@@ -588,7 +588,7 @@ static void put_io_u(struct thread_data *td, struct io_u *io_u)
 
 #define queue_full(td)	(list_empty(&(td)->io_u_freelist))
 
-static struct io_u *__get_io_u(struct thread_data *td)
+struct io_u *__get_io_u(struct thread_data *td)
 {
 	struct io_u *io_u;
 
@@ -1113,6 +1113,7 @@ static int init_io_u(struct thread_data *td)
 		INIT_LIST_HEAD(&io_u->list);
 
 		io_u->buf = p + td->max_bs * i;
+		io_u->index = i;
 		list_add(&io_u->list, &td->io_u_freelist);
 	}
 
@@ -1208,6 +1209,7 @@ static int file_size(struct thread_data *td)
 	if (!td->file_size || td->file_size > td->real_file_size)
 		td->file_size = td->real_file_size;
 
+	td->file_size -= td->file_offset;
 	return 0;
 }
 
@@ -1230,23 +1232,26 @@ static int bdev_size(struct thread_data *td)
 	if (!td->file_size || td->file_size > td->real_file_size)
 		td->file_size = td->real_file_size;
 
+	td->file_size -= td->file_offset;
 	return 0;
 }
 
 static int get_file_size(struct thread_data *td)
 {
-	int ret;
+	int ret = 0;
 
 	if (td->filetype == FIO_TYPE_FILE)
 		ret = file_size(td);
-	else
+	else if (td->filetype == FIO_TYPE_BD)
 		ret = bdev_size(td);
+	else
+		td->real_file_size = -1;
 
 	if (ret)
 		return ret;
 
-	if (td->file_offset + td->file_size > td->real_file_size) {
-		fprintf(stderr, "Client%d: offset extends end (%Lu > %Lu)\n", td->thread_number, td->file_offset + td->file_size, td->real_file_size);
+	if (td->file_offset > td->real_file_size) {
+		fprintf(stderr, "Client%d: offset extends end (%Lu > %Lu)\n", td->thread_number, td->file_offset, td->real_file_size);
 		return 1;
 	}
 
@@ -1347,7 +1352,7 @@ static int setup_file(struct thread_data *td)
 		if (create_file(td, td->file_size, 0))
 			return 1;
 	} else if (td->filetype == FIO_TYPE_FILE) {
-		if (st.st_size < td->file_size) {
+		if (st.st_size < (off_t) td->file_size) {
 			if (create_file(td, td->file_size - st.st_size, 1))
 				return 1;
 		}
@@ -1356,9 +1361,14 @@ static int setup_file(struct thread_data *td)
 	if (td->odirect)
 		flags |= O_DIRECT;
 
-	if (td_read(td))
-		td->fd = open(td->file_name, flags | O_RDONLY);
-	else {
+	if (td_read(td)) {
+		if (td->filetype == FIO_TYPE_CHAR)
+			flags |= O_RDWR;
+		else
+			flags |= O_RDONLY;
+
+		td->fd = open(td->file_name, flags);
+	} else {
 		if (td->filetype == FIO_TYPE_FILE) {
 			if (!td->overwrite)
 				flags |= O_TRUNC;
@@ -1599,17 +1609,23 @@ static void init_disk_util(struct thread_data *td)
 		return;
 
 	/*
-	 * if this is inside a partition dir, jump back to parent
+	 * for md/dm, there's no queue dir. we already have the right place
 	 */
-	sprintf(tmp, "%s/queue", foo);
+	sprintf(tmp, "%s/stat", foo);
 	if (stat(tmp, &st)) {
-		p = dirname(foo);
-		sprintf(tmp, "%s/queue", p);
+		/*
+		 * if this is inside a partition dir, jump back to parent
+		 */
+		sprintf(tmp, "%s/queue", foo);
 		if (stat(tmp, &st)) {
-			fprintf(stderr, "unknown sysfs layout\n");
-			return;
+			p = dirname(foo);
+			sprintf(tmp, "%s/queue", p);
+			if (stat(tmp, &st)) {
+				fprintf(stderr, "unknown sysfs layout\n");
+				return;
+			}
+			sprintf(foo, "%s", p);
 		}
-		sprintf(foo, "%s", p);
 	}
 
 	disk_util_add(dev, foo);
@@ -1915,14 +1931,70 @@ static void eta_to_str(char *str, int eta_sec)
 	str += sprintf(str, "%02ds", s);
 }
 
+static int thread_eta(struct thread_data *td, unsigned long elapsed)
+{
+	unsigned long long bytes_total, bytes_done;
+	unsigned int eta_sec = 0;
+
+	bytes_total = td->total_io_size;
+	if (td->verify)
+		bytes_total <<= 1;
+	if (td->zone_size && td->zone_skip)
+		bytes_total /= (td->zone_skip / td->zone_size);
+
+	if (td->runstate == TD_RUNNING || td->runstate == TD_VERIFYING) {
+		double perc;
+		bytes_done = td->io_bytes[DDIR_READ] + td->io_bytes[DDIR_WRITE];
+		perc = (double) bytes_done / (double) bytes_total;
+		if (perc > 1.0)
+			perc = 1.0;
+
+		eta_sec = (elapsed * (1.0 / perc)) - elapsed;
+
+		if (eta_sec > (td->timeout - elapsed))
+			eta_sec = td->timeout - elapsed;
+	} else if (td->runstate == TD_NOT_CREATED || td->runstate == TD_CREATED) {
+		int t_eta = 0, r_eta = 0;
+
+		/*
+		 * We can only guess - assume it'll run the full timeout
+		 * if given, otherwise assume it'll run at the specified rate.
+		 */
+		if (td->timeout)
+			t_eta = td->timeout + td->start_delay - elapsed;
+		if (td->rate) {
+			r_eta = (bytes_total / 1024) / td->rate;
+			r_eta += td->start_delay - elapsed;
+		}
+
+		if (r_eta && t_eta)
+			eta_sec = min(r_eta, t_eta);
+		else if (r_eta)
+			eta_sec = r_eta;
+		else if (t_eta)
+			eta_sec = t_eta;
+		else
+			eta_sec = INT_MAX;
+	} else {
+		/*
+		 * thread is already done
+		 */
+		eta_sec = 0;
+	}
+
+	return eta_sec;
+}
+
 static void print_thread_status(void)
 {
-	unsigned long long bytes_done, bytes_total;
-	int i, nr_running, t_rate, m_rate, eta_sec;
+	unsigned long elapsed = time_since_now(&genesis);
+	int i, nr_running, t_rate, m_rate, *eta_secs, eta_sec;
 	char eta_str[32];
-	double perc;
+	double perc = 0.0;
+
+	eta_secs = malloc(thread_number * sizeof(int));
+	memset(eta_secs, 0, thread_number * sizeof(int));
 
-	bytes_done = bytes_total = 0;
 	nr_running = t_rate = m_rate = 0;
 	for (i = 0; i < thread_number; i++) {
 		struct thread_data *td = &threads[i];
@@ -1933,45 +2005,45 @@ static void print_thread_status(void)
 			m_rate += td->ratemin;
 		}
 
-		bytes_total += td->total_io_size;
-		if (td->verify)
-			bytes_total += td->total_io_size;
-
-		if (td->zone_size && td->zone_skip)
-			bytes_total /= (td->zone_skip / td->zone_size);
-	
-		bytes_done += td->io_bytes[DDIR_READ] +td->io_bytes[DDIR_WRITE];
+		if (elapsed >= 3)
+			eta_secs[i] = thread_eta(td, elapsed);
 
 		check_str_update(td);
 	}
 
-	perc = 0;
-	eta_sec = 0;
-	if (bytes_total && bytes_done) {
-		unsigned long runtime;
+	if (exitall_on_terminate)
+		eta_sec = INT_MAX;
+	else
+		eta_sec = 0;
 
-		perc = (double) bytes_done / (double) bytes_total;
-		if (perc > 1.0)
-			perc = 1.0;
-
-		runtime = time_since_now(&genesis);
-		if (runtime >= 5) {
-			memset(eta_str, 0, sizeof(eta_str));
-			eta_sec = (runtime * (1.0 / perc)) - runtime;
-			eta_to_str(eta_str, eta_sec);
+	for (i = 0; i < thread_number; i++) {
+		if (exitall_on_terminate) {
+			if (eta_secs[i] < eta_sec)
+				eta_sec = eta_secs[i];
+		} else {
+			if (eta_secs[i] > eta_sec)
+				eta_sec = eta_secs[i];
 		}
+	}
 
-		perc *= 100.0;
+	if (eta_sec == INT_MAX)
+		eta_sec = 0;
+
+	if (eta_sec) {
+		perc = (double) elapsed / (double) (elapsed + eta_sec);
+		eta_to_str(eta_str, eta_sec);
 	}
 
 	printf("Threads now running (%d)", nr_running);
 	if (m_rate || t_rate)
 		printf(", commitrate %d/%dKiB/sec", t_rate, m_rate);
-	printf(": [%s] [%3.2f%% done]", run_str, perc);
-	if (eta_sec)
-		printf(" [eta %s]", eta_str);
+	if (eta_sec) {
+		perc *= 100.0;
+		printf(": [%s] [%3.2f%% done] [eta %s]", run_str, perc,eta_str);
+	}
 	printf("\r");
 	fflush(stdout);
+	free(eta_secs);
 }
 
 static void reap_threads(int *nr_running, int *t_rate, int *m_rate)
